@@ -1,19 +1,21 @@
 # coding=utf-8
 from dataclasses import dataclass, field
 from datetime import date, datetime
-from typing import Any, cast, Dict, List, Optional, Tuple, Union
+from typing import Any, cast, Dict, List, Optional, Tuple, TypedDict, Union
 
 import yaml
 from dateutil.tz import tzutc
 from networkx import ancestors, DiGraph
+from zuper_commons.types import ZException, ZValueError
 from zuper_ipce import ipce_from_object, object_from_ipce
 
-from .challenges_constants import ChallengesConstants, JobStatusString, StepName
-from .cmd_submit_build import parse_complete_tag
+from .challenges_constants import ChallengesConstants
 from .exceptions import InvalidConfiguration
+from .types import ChallengeName, JobStatusString, ServiceName, StepName
 from .utils import indent, safe_yaml_dump, wrap_config_reader2
 
 __all__ = [
+    "EvaluationParametersDict",
     "ChallengeDescription",
     "EvaluationParameters",
     "ChallengesConstants",
@@ -31,10 +33,16 @@ __all__ = [
     "STATE_FAILED",
     "STATE_SUCCESS",
     "ALLOWED_CONDITION_TRIGGERS",
+    "DepBetter",
+    "DepScore",
+    "from_steps_transitions",
+    "Scoring_as_dict",
+    "steps_from_transitions",
+    "Transition",
 ]
 
 
-class InvalidChallengeDescription(Exception):
+class InvalidChallengeDescription(ZException):
     pass
 
 
@@ -88,11 +96,18 @@ class PortDefinition:
     internal: int
 
 
+class ServiceDefinitionDict(TypedDict):
+    image: Optional[str]
+    build: Optional[Build]
+    environment: Dict[str, Any]
+    ports: List[PortDefinition]
+
+
 @dataclass
 class ServiceDefinition:
     image: Optional[str]
     build: Optional[Build]
-    environment: Dict[str, Any]
+    environment: Dict[str, str]
     ports: List[PortDefinition]
 
     def __repr__(self):
@@ -100,51 +115,50 @@ class ServiceDefinition:
 
     def equivalent(self, other):
         if self.image != ChallengesConstants.SUBMISSION_CONTAINER_TAG:
+            from duckietown_build_utils import parse_complete_tag, DockerCompleteImageName  # FIXME
+
             br2 = parse_complete_tag(other.image)
 
             try:
-                br1 = parse_complete_tag(self.image)
+                br1 = parse_complete_tag(DockerCompleteImageName(self.image))
             except ValueError as e:
                 msg = "Could not even parse mine"
                 raise NotEquivalent(msg) from e
 
             if br1.digest is None or br2.digest is None:
-                msg = "No digest information, assuming different.\nself: %s\nother: %s" % (br1, br2)
-                raise NotEquivalent(msg)
+                msg = "No digest information, assuming different."
+                raise NotEquivalent(msg, br1=br1, br2=br2)
             if br1.digest != br2.digest:
-                msg = "Different digests:\n\n  %s\n\n  %s" % (br1, br2)
-                raise NotEquivalent(msg)
+                msg = "Different digests"
+                raise NotEquivalent(msg, br1=br1, br2=br2)
             if self.ports != other.ports:
-                msg = "Different digests:\n\n  %s\n\n  %s" % (br1, br2)
-                raise NotEquivalent(msg)
+                msg = "Different ports"
+                raise NotEquivalent(msg, mine=self.ports, other=self.ports)
 
         if self.environment != other.environment:
-            msg = "Different environments:\n\n %s\n\n  %s" % (self.environment, other.environment,)
-            raise NotEquivalent(msg)
+            msg = "Different environments"
+            raise NotEquivalent(msg, mine=self.environment, other=other.environment)
 
     # noinspection PyArgumentList
     @classmethod
     @wrap_config_reader2
-    def from_yaml(cls, d0):
-        image = d0.pop("image", None)
-        environment = d0.pop("environment", {})
+    def from_yaml(cls, d0: ServiceDefinitionDict) -> "ServiceDefinition":
+        image = d0.get("image", None)
+        environment = d0.get("environment", {})
         if environment is None:
             environment = {}
 
-        if "build" in d0:
-            build = d0.pop("build")
-            if build is not None:
-                build = Build.from_yaml(build)
-        else:
-            build = None
+        build = d0.get("build", None)
+        if build is not None:
+            build = Build.from_yaml(build)
 
         if build and image:
             msg = 'Cannot specify both "build" and "image".'
             raise ValueError(msg)
 
-        image_digest = d0.pop("image_digest", None)
-        if image_digest:
-            image = f"{image}@{image_digest}"
+        # image_digest = d0.get("image_digest", None)
+        # if image_digest:
+        #     image = f"{image}@{image_digest}"
 
         for k, v in list(environment.items()):
             if "-" in k:
@@ -162,7 +176,7 @@ class ServiceDefinition:
             else:
                 msg = f'The type {type(v).__name__} is not allowed for environment variable "{k}".'
                 raise InvalidConfiguration(msg)
-        ports_ = d0.pop("ports", [])
+        ports_ = d0.get("ports", [])
         ports = []
         for s in ports_:
             if isinstance(s, int):
@@ -206,6 +220,11 @@ class ServiceDefinition:
         return res
 
 
+class EvaluationParametersDict(TypedDict):
+    version: str
+    services: Dict[str, ServiceDefinitionDict]
+
+
 @dataclass
 class EvaluationParameters:
     """
@@ -226,7 +245,7 @@ class EvaluationParameters:
     """
 
     version: str
-    services: Dict[str, ServiceDefinition]
+    services: Dict[ServiceName, ServiceDefinition]
 
     def __init__(self, version, services):
         self.version = version
@@ -235,18 +254,18 @@ class EvaluationParameters:
     # noinspection PyArgumentList
     @classmethod
     @wrap_config_reader2
-    def from_yaml(cls, d):
+    def from_yaml(cls, d: EvaluationParametersDict) -> "EvaluationParameters":
 
-        services_ = d.pop("services")
+        services_ = d.get("services")
         if not isinstance(services_, dict):
-            msg = "Expected dict got %s" % services_.__repr__()
-            raise ValueError(msg)
+            msg = "Expected dict"
+            raise ZValueError(msg, got=services_)
 
         if not services_:
             msg = "No services described."
             raise ValueError(msg)
 
-        version = d.pop("version", "3")
+        version = d.get("version", "3")
 
         services = {}
         for k, v in services_.items():
@@ -274,16 +293,18 @@ class EvaluationParameters:
         services = dict([(k, v.as_dict()) for k, v in self.services.items()])
         return dict(version=self.version, services=services)
 
-    def equivalent(self, other):
+    def equivalent(self, other: "EvaluationParameters"):
         if set(other.services) != set(self.services):
             msg = "Different set of services."
             raise NotEquivalent(msg)
         for s in other.services:
+            mine = self.services[s]
+            its = other.services[s]
             try:
-                self.services[s].equivalent(other.services[s])
+                mine.equivalent(its)
             except NotEquivalent as e:
-                msg = "Service %r differs:\n\n%s" % (s, indent(e, "  "))
-                raise NotEquivalent(msg)
+                msg = f"Service {s} differs"
+                raise NotEquivalent(msg, mine=mine, its=its) from e
 
 
 @dataclass
@@ -309,7 +330,7 @@ class ChallengeStep:
     # noinspection PyArgumentList
     @classmethod
     @wrap_config_reader2
-    def from_yaml(cls, data, name):
+    def from_yaml(cls, data, name) -> "ChallengeStep":
         title = data.pop("title")
         description = data.pop("description")
         evaluation_parameters = EvaluationParameters.from_yaml(data.pop("evaluation_parameters"))
@@ -327,7 +348,7 @@ class ChallengeStep:
         )
 
 
-class NotEquivalent(Exception):
+class NotEquivalent(ZException):
     pass
 
 
@@ -339,30 +360,20 @@ def nice_repr(x):
 # Transition = namedtuple("Transition", "first condition second")
 @dataclass
 class Transition:
-    first: str
+    first: StepName
     condition: str
-    second: str
+    second: StepName
 
 
-class InvalidSteps(Exception):
+class InvalidSteps(ZException):
     pass
 
 
 @dataclass(repr=False)
 class ChallengeTransitions:
-    steps: List[str]
+    steps: List[StepName]
     transitions: List[Transition]
 
-    #
-    # def __init__(self, transitions, steps):
-    #     self.transitions = []
-    #     self.steps = steps
-    #     for first, condition, second in self.transitions:
-    #         assert first == STATE_START or first in self.steps, first
-    #         assert second in [STATE_ERROR, STATE_FAILED, STATE_SUCCESS] or second in self.steps, second
-    #         assert condition in ALLOWED_CONDITION_TRIGGERS, condition
-    #         self.transitions.append(Transition(first, condition, second))
-
     def as_list(self):
         res = []
         for transition in self.transitions:
@@ -400,7 +411,7 @@ class ChallengeTransitions:
             G.add_edge(t.first, t.second)
         return G
 
-    def get_precs(self, x: str) -> List[str]:
+    def get_precs(self, x: StepName) -> List[StepName]:
         G = self.get_graph()
         res = list(ancestors(G, x))
         # print('precs of %s: %s' % (x, res))
@@ -422,20 +433,21 @@ class ChallengeTransitions:
 
         """
         CS = ChallengesConstants
-        # dclogger.info('Received status = %s' % status)
+        # logger.info('Received status = %s' % status)
         assert isinstance(status, dict)
         assert STATE_START in status
+        # noinspection PyTypeChecker
         assert status[STATE_START] == CS.STATUS_JOB_SUCCESS
         status = dict(**status)
         for k, ks in list(status.items()):
             if k != STATE_START and k not in self.steps:
                 # msg = "Ignoring invalid step %s -> %s" % (k, ks)
-                # dclogger.error(msg)
+                # logger.error(msg)
                 status.pop(k)
 
             if ks not in ChallengesConstants.ALLOWED_JOB_STATUS:
                 # msg = f"Ignoring invalid step {k} -> {ks!r}"
-                # dclogger.error(msg)
+                # logger.error(msg)
                 status.pop(k)
 
             # timeout or aborted or host error = like it never happened
@@ -449,14 +461,14 @@ class ChallengeTransitions:
 
         # make sure that the steps in which they depend are ok
 
-        def predecessors_success(_: str) -> bool:
+        def predecessors_success(_: StepName) -> bool:
             precs = self.get_precs(_)
             its_age = step2age.get(_, -1) if step2age else 0
             for k2 in precs:
                 pred_age = step2age.get(k2, -1) if step2age else 0
-                # dclogger.debug('%s %s %s %s' % (_, its_age, k2, pred_age))
+                # logger.debug('%s %s %s %s' % (_, its_age, k2, pred_age))
                 if pred_age > its_age:
-                    # dclogger.debug('Its depedency is younger')
+                    # logger.debug('Its depedency is younger')
                     return False
                 if k2 not in status or status[k2] != CS.STATUS_JOB_SUCCESS:
                     return False
@@ -466,7 +478,7 @@ class ChallengeTransitions:
         outcomes = set()
         for t in self.transitions:
             if t.first in status and status[t.first] == t.condition and predecessors_success(t.first):
-                # dclogger.debug('Transition %s is activated' % str(t))
+                # logger.debug('Transition %s is activated' % str(t))
 
                 like_it_does_not_exist = [ChallengesConstants.STATUS_JOB_ABORTED]
                 if (
@@ -474,11 +486,11 @@ class ChallengeTransitions:
                     and status[t.second] not in like_it_does_not_exist
                     and predecessors_success(t.second)
                 ):
-                    # dclogger.debug('Second %s already activated (and in %s)' % (t.second, status[t.second]))
+                    # logger.debug('Second %s already activated (and in %s)' % (t.second, status[t.second]))
                     pass
                 else:
                     if t.second in [STATE_ERROR, STATE_FAILED, STATE_SUCCESS]:
-                        # dclogger.debug('Finishing here')
+                        # logger.debug('Finishing here')
                         outcomes.add(t.second.lower())
                         # return True, t.second.lower(), []
                     else:
@@ -490,11 +502,11 @@ class ChallengeTransitions:
             complete = False
             outcome = None
 
-        # dclogger.debug('Incomplete; need to do: %s' % to_activate)
+        # logger.debug('Incomplete; need to do: %s' % to_activate)
         return complete, outcome, to_activate
 
 
-def steps_from_transitions(transitions: List[List[str]]) -> List[str]:
+def steps_from_transitions(transitions: List[List[str]]) -> List[StepName]:
     steps = []
     for first, _, second in transitions:
         if first not in [STATE_ERROR, STATE_FAILED, STATE_SUCCESS]:
@@ -504,7 +516,7 @@ def steps_from_transitions(transitions: List[List[str]]) -> List[str]:
     return steps
 
 
-def from_steps_transitions(steps: List[str], transitions_str: List[List[str]]) -> ChallengeTransitions:
+def from_steps_transitions(steps: List[StepName], transitions_str: List[List[str]]) -> ChallengeTransitions:
     transitions = []
     for first, condition, second in transitions_str:
         assert first == STATE_START or first in steps, first
@@ -538,14 +550,14 @@ class Score:
 
     def __post_init__(self):
         if self.order not in Score.ALLOWED:
-            msg = "Invalid value %s" % self.order
-            raise ValueError(msg)
+            msg = "Invalid value"
+            raise ZValueError(msg, order=self.order, allowed=Score.ALLOWED)
 
         if self.discretization is not None:
             discretization = float(self.discretization)
             if discretization <= 0:
-                msg = "Need a strictly positive discretization: %s" % discretization
-                raise ValueError(msg)
+                msg = "Need a strictly positive discretization"
+                raise ZValueError(msg, discretization=discretization)
 
     def __repr__(self):
         return nice_repr(self)
@@ -675,28 +687,28 @@ class ChallengeDependency:
 
 @dataclass
 class ChallengeDescription:
-    name: str
+    name: ChallengeName
     title: str
     description: str
     protocol: str
     date_open: datetime
     date_close: datetime
-    steps: Dict[str, ChallengeStep]
+    steps: Dict[StepName, ChallengeStep]
     transitions: List[List[str]]  # 3 elements each
 
     scoring: Scoring
     tags: List[str]
 
-    dependencies: Dict[str, ChallengeDependency]
+    dependencies: Dict[ChallengeName, ChallengeDependency]
 
     ct: ChallengeTransitions
-    closure: Optional[List[str]] = field(default_factory=list)
+    closure: Optional[List[ChallengeName]] = field(default_factory=list)
 
     def __post_init__(self):
 
         # check_isinstance(self.date_open, datetime)
         # check_isinstance(self.date_close, datetime)
-        # dclogger.info(f'received {self.date_open.tzinfo} {id(self.date_open)}   {self.date_close.tzinfo}'
+        # logger.info(f'received {self.date_open.tzinfo} {id(self.date_open)}   {self.date_close.tzinfo}'
         #               f' {id(self.date_close)}  ')
         if self.date_open.tzinfo is None:
             raise ValueError(self.date_open)
@@ -787,6 +799,7 @@ class ChallengeDescription:
         data["transitions"] = []
         for t in self.ct.transitions:
             tt = [t.first, t.condition, t.second]
+            # noinspection PyTypeChecker
             data["transitions"].append(tt)
         steps = {}
         for k, v in self.steps.items():
@@ -851,7 +864,7 @@ class SubmissionDescription:
     challenge_names: Optional[List[str]]
     protocols: List[str]
     user_label: Optional[str]
-    user_metadata: dict
+    user_metadata: Dict[str, object]
     description: Optional[str]
 
     def __post_init_(self):
