@@ -1,10 +1,9 @@
 # coding=utf-8
-from typing import cast, ClassVar, NewType
+from typing import cast, ClassVar
 
-__all__ = ["JobStatusString", "StepName", "ChallengesConstants"]
+__all__ = ["ChallengesConstants"]
 
-JobStatusString = NewType("JobStatusString", str)
-StepName = NewType("StepName", str)
+from .types import JobStatusString
 
 
 class ChallengesConstants:
@@ -13,9 +12,7 @@ class ChallengesConstants:
     STATUS_JOB_EVALUATION: ClassVar[JobStatusString] = cast(JobStatusString, "evaluating")
     STATUS_JOB_FAILED: ClassVar[JobStatusString] = cast(JobStatusString, "failed")
     STATUS_JOB_ERROR: ClassVar[JobStatusString] = cast(JobStatusString, "error")  # evaluation failed
-    STATUS_JOB_HOST_ERROR: ClassVar[JobStatusString] = cast(
-        JobStatusString, "host-error"
-    )  # evaluation failed
+    STATUS_JOB_HOST_ERROR: ClassVar[JobStatusString] = cast(JobStatusString, "host-error")
     STATUS_JOB_SUCCESS: ClassVar[JobStatusString] = cast(JobStatusString, "success")
     STATUS_JOB_ABORTED: ClassVar[JobStatusString] = cast(JobStatusString, "aborted")
 
